@@ -10,16 +10,17 @@ independent data models in python. This gives us:
 
 """
 
-import arrow
+from __future__ import annotations
+
 import copy
 import json
 import logging
 import re
-import typing
 import sys
-
-from dateutil.tz import tzutc
 from datetime import datetime
+from typing import Any, Dict, Mapping, Tuple, Union
+import arrow
+from dateutil.tz import tzutc
 
 from assemblyline.common import forge
 from assemblyline.common.dict_utils import recursive_update
@@ -64,8 +65,8 @@ PROCESSOR_REGEX = r"^x(64|86)$"
 logger = logging.getLogger('assemblyline.odm')
 
 
-def flat_to_nested(data: dict):
-    sub_data = {}
+def flat_to_nested(data: dict[str, Any]) -> dict[str, Any]:
+    sub_data: dict[str, Any] = {}
     nested_keys = []
     for key, value in data.items():
         if '.' in key:
@@ -88,12 +89,8 @@ class KeyMaskException(KeyError):
     pass
 
 
-class UndefinedFunction(Exception):
-    pass
-
-
 class _Field:
-    def __init__(self, name=None, index=None, store=None, copyto=None, default=None):
+    def __init__(self, name=None, index=None, store=None, copyto=None, default=None, description=None):
         self.index = index
         self.store = store
         self.multivalued = False
@@ -107,6 +104,7 @@ class _Field:
         self.parent_name = None
         self.getter_function = None
         self.setter_function = None
+        self.description = description
 
         self.default = default
         self.default_set = default is not None
@@ -170,24 +168,14 @@ class _Field:
         return {'': self}
 
     def check(self, value, **kwargs):
-        raise UndefinedFunction("This function is not defined in the default field. "
-                                "Each fields has to have their own definition")
+        raise NotImplementedError("This function is not defined in the default field. "
+                                  "Each fields has to have their own definition")
 
 
 class _DeletedField:
     pass
 
 
-class Any(_Field):
-    def __init__(self, *args, **kwargs):
-        kwargs['index'] = False
-        kwargs['store'] = False
-        super().__init__(*args, **kwargs)
-
-    def check(self, value, **kwargs):
-        return value
-
-
 class Date(_Field):
     """A field storing a datetime value."""
 
@@ -211,6 +199,19 @@ class Boolean(_Field):
         return bool(value)
 
 
+class Json(_Field):
+    """
+    A field storing serializeable structure with their JSON encoded representations.
+
+    Examples: metadata
+    """
+
+    def check(self, value, **kwargs):
+        if not isinstance(value, str):
+            return json.dumps(value)
+        return value
+
+
 class Keyword(_Field):
     """
     A field storing a short string with a technical interpretation.
@@ -236,6 +237,55 @@ class Keyword(_Field):
         return str(value)
 
 
+class EmptyableKeyword(_Field):
+    """
+    A keyword which allow to differentiate between empty and None values.
+    """
+
+    def check(self, value, **kwargs):
+        # We have a special case for bytes here due to how often strings and bytes
+        # get mixed up in python apis
+        if isinstance(value, bytes):
+            raise ValueError(f"[{self.name or self.parent_name}] EmptyableKeyword doesn't accept bytes values")
+
+        if value is None and self.default_set:
+            value = self.default
+
+        if value is None:
+            return None
+
+        return str(value)
+
+
+class UpperKeyword(Keyword):
+    """
+    A field storing a short uppercase string with a technical interpretation.
+    """
+
+    def check(self, value, **kwargs):
+        kw_val = super().check(value, **kwargs)
+
+        if kw_val is None:
+            return None
+
+        return kw_val.upper()
+
+
+class Any(Keyword):
+    """
+    A field that can hold any value whatsoever but which is stored as a
+    Keyword in the datastore index
+    """
+
+    def __init__(self, *args, **kwargs):
+        kwargs['index'] = False
+        kwargs['store'] = False
+        super().__init__(*args, **kwargs)
+
+    def check(self, value, **_):
+        return value
+
+
 class ValidatedKeyword(Keyword):
     """
     Keyword field which the values are validated by a regular expression
@@ -684,7 +734,7 @@ class FlattenedListObject(Mapping):
     """A field storing a flattened object"""
 
     def __init__(self, **kwargs):
-        super().__init__(List(Keyword()), **kwargs)
+        super().__init__(List(Json()), **kwargs)
 
     def check(self, value, **kwargs):
         return TypedMapping(self.child_type, self.index, self.store, FLATTENED_OBJECT_SANITIZER, **value)
@@ -701,7 +751,7 @@ class FlattenedObject(Mapping):
     """A field storing a flattened object"""
 
     def __init__(self, **kwargs):
-        super().__init__(Keyword(), **kwargs)
+        super().__init__(Json(), **kwargs)
 
     def check(self, value, **kwargs):
         return TypedMapping(self.child_type, self.index, self.store, FLATTENED_OBJECT_SANITIZER, **value)
@@ -719,10 +769,10 @@ class Compound(_Field):
         super().__init__(**kwargs)
         self.child_type = field_type
 
-    def check(self, value, mask=None, ignore_extra_values=False, **kwargs):
+    def check(self, value, mask=None, ignore_extra_values=False, extra_fields={}, **kwargs):
         if isinstance(value, self.child_type):
             return value
-        return self.child_type(value, mask=mask, ignore_extra_values=ignore_extra_values)
+        return self.child_type(value, mask=mask, ignore_extra_values=ignore_extra_values, extra_fields=extra_fields)
 
     def fields(self):
         out = dict()
@@ -758,11 +808,7 @@ class Optional(_Field):
 
 class Model:
     @classmethod
-    def name(cls):
-        return cls.__name__
-
-    @classmethod
-    def fields(cls, skip_mappings=False) -> typing.Mapping[str, _Field]:
+    def fields(cls, skip_mappings=False) -> Mapping[str, _Field]:
         """
         Describe the elements of the model.
 
@@ -799,9 +845,6 @@ class Model:
             if skip_mappings and isinstance(sub_field, Mapping):
                 continue
 
-            elif isinstance(sub_field, Any):
-                continue
-
             elif isinstance(sub_field, (List, Optional, Compound)) and sub_name != "":
                 out.update(Model._recurse_fields(".".join([name, sub_name]), sub_field.child_type,
                                                  show_compound, skip_mappings,
@@ -819,7 +862,7 @@ class Model:
         return out
 
     @classmethod
-    def flat_fields(cls, show_compound=False, skip_mappings=False) -> typing.Mapping[str, _Field]:
+    def flat_fields(cls, show_compound=False, skip_mappings=False) -> Mapping[str, _Field]:
         """
         Describe the elements of the model.
 
@@ -834,16 +877,98 @@ class Model:
             if isinstance(field, _Field):
                 if skip_mappings and isinstance(field, Mapping):
                     continue
-                if isinstance(field, Any):
-                    continue
                 out.update(Model._recurse_fields(name, field, show_compound, skip_mappings,
                                                  multivalued=isinstance(field, List)))
         return out
 
+    @classmethod
+    def markdown(cls, toc_depth=1, defaults=None) -> Union[str, Dict]:
+        markdown_content = "[comment]: # (AUTOGENERATED MARKDOWN CONTENT. UPDATES TO ODM DOCUMENTATION SHOULD BE DONE THROUGH ASSEMBLYLINE-BASE REPO!)\n"
+
+        # Header
+        markdown_content += f"{'#'*toc_depth} {cls.__name__}\n> {cls.__description}\n\n"
+
+        # Table
+        table = "| Field | Type | Description | Required | Default |\n| :--- | :--- | :--- | :--- | :--- |\n"
+
+        # Determine the type of Field we're dealing with
+        # if possible return the Model class if wrapped in Compound
+        def get_type(field_class: _Field) -> Tuple(str, Model):
+            if field_class.__class__ == Optional:
+                return get_type(field_class.child_type)
+            elif field_class.__class__ == Compound:
+                module_path = field_class.child_type.__module__
+                if '/' in module_path:
+                    module_path = module_path[module_path.index('/odm'):-3]
+                else:
+                    module_path = module_path.split('assemblyline')[1].replace('.', '/')
+                name = field_class.child_type.__name__
+                return f"[{name}]({module_path}/#{name.lower()})", field_class.child_type
+            elif field_class.__class__ in [Mapping, List]:
+                child_type, child_class = field_class.child_type.__class__.__name__, field_class.child_type.__class__
+                if field_class.child_type.__class__ == Compound:
+                    child_type, child_class = get_type(field_class.child_type)
+                return f"{field_class.__class__.__name__} [{child_type}]", child_class
+            elif field_class.__class__.__name__ == 'type':
+                return field_class.__name__, None
+
+            return field_class.__class__.__name__, None
+
+        model_deps = list()
+        for field, info in cls.fields().items():
+            field_type, field_class = get_type(info)
+
+            # If the field is in fact a model class, generate markdown and append to list
+            if field_class and issubclass(field_class, Model) and field_class.__module__ == cls.__module__:
+                model_deps.append(field_class.markdown(toc_depth=toc_depth+1, defaults=info.default))
+
+            # Field description
+            description = info.description
+
+            # If field type is Enum, then show the possible values that can be used in the description
+            if field_type == "Enum":
+                values = info.child_type.values if info.__class__ != Enum else info.values
+                none_value = False
+                if None in values:
+                    none_value = True
+                    values.remove(None)
+
+                values = [f'"{v}"' if v else str(v) for v in sorted(values)]
+                values.append("None") if none_value else None
+                description = f'{description}<br>Values:<br>`{", ".join(values)}`'
+
+            # Is this a required field?
+            required = ":material-checkbox-marked-outline: Yes" if info.__class__ != Optional else ":material-minus-box-outline: Optional"
+
+            # Determine the correct default values to display
+            default = f"`{info.default}`"
+            # If the field is a model, then provide a link to that documentation
+            if field_class and issubclass(field_class, Model) and isinstance(info.default, dict):
+                ref_link = field_type[field_type.index('('):field_type.index(')')+1]
+                default = f"See [{field_class.__name__}]{ref_link} for more details."
+
+            # Handle how to display values from provided defaults (different from field defaults)
+            elif isinstance(defaults, dict):
+                val = defaults.get(field, {})
+                default = f"`{val if not isinstance(val, dict) else info.default}`"
+            elif isinstance(defaults, list):
+                default = f'`{defaults}`'
+            row = f"| {field} | {field_type} | {description} | {required} | {default} |\n"
+            table += row
+
+        markdown_content += table + "\n\n"
+        # Display model dependencies in order
+        for markdown in sorted(model_deps):
+            markdown_content += markdown
+
+        return markdown_content
+
     # Allow attribute assignment by default in the constructor until it is removed
     __frozen = False
+    # Descriptions of the model should be class-accessible only for markdown()
+    __description = None
 
-    def __init__(self, data: dict = None, mask: list = None, docid=None, ignore_extra_values=True):
+    def __init__(self, data: dict = None, mask: list = None, docid=None, ignore_extra_values=True, extra_fields={}):
         if data is None:
             data = {}
         if not hasattr(data, 'items'):
@@ -876,10 +1001,11 @@ class Model:
 
         # Check to make sure we can use all the data we are given
         unused_keys = set(data.keys()) - set(fields.keys()) - BANNED_FIELDS
+        extra_keys = set(extra_fields.keys()) - set(data.keys())
         if unused_keys and not ignore_extra_values:
             raise ValueError(f"'{self.__class__.__name__}' object was created with invalid parameters: "
                              f"{', '.join(unused_keys)}")
-        if unused_keys and ignore_extra_values:
+        if unused_keys and ignore_extra_values and mask is None:
             logger.warning(
                 f"The following parameters where ignored from object "
                 f"'{self.__class__.__name__}': {', '.join(unused_keys)}")
@@ -889,6 +1015,8 @@ class Model:
             params = {"ignore_extra_values": ignore_extra_values}
             if name in mask_map and mask_map[name]:
                 params['mask'] = mask_map[name]
+            if name in extra_fields and extra_fields[name]:
+                params['extra_fields'] = extra_fields[name]
 
             try:
                 value = data[name]
@@ -900,6 +1028,9 @@ class Model:
 
             self._odm_py_obj[name] = field_type.check(value, **params)
 
+        for key in extra_keys:
+            self._odm_py_obj[key] = Any().check(extra_fields[key])
+
         # Since the layout of model objects should be fixed, don't allow any further
         # attribute assignment
         self.__frozen = True
@@ -910,7 +1041,7 @@ class Model:
 
         fields = self.fields()
         for key, value in self._odm_py_obj.items():
-            field_type = fields[key]
+            field_type = fields.get(key, Any)
             if value is not None or (value is None and field_type.default_set):
                 if strip_null and value is None:
                     continue
@@ -922,7 +1053,7 @@ class Model:
                 elif isinstance(value, TypedMapping):
                     out[key] = {k: v.as_primitives(strip_null=strip_null)
                                 if isinstance(v, Model) else v for k, v in value.items()}
-                elif isinstance(value, (List, TypedList)):
+                elif isinstance(value, TypedList):
                     out[key] = [v.as_primitives(strip_null=strip_null)
                                 if isinstance(v, Model) else v for v in value]
                 elif isinstance(value, ClassificationObject):
@@ -981,7 +1112,7 @@ class Model:
         return object.__setattr__(self, name, value)
 
 
-def model(index=None, store=None):
+def model(index=None, store=None, description=None):
     def recursive_set_name(field, name, to_parent=False):
         if not to_parent:
             field.name = name
@@ -995,6 +1126,7 @@ def model(index=None, store=None):
 
     """Decorator to create model objects."""
     def _finish_model(cls):
+        cls._Model__description = description
         for name, field_data in cls.fields().items():
             if not FIELD_SANITIZER.match(name) or name in BANNED_FIELDS:
                 raise ValueError(f"Illegal variable name: {name}")
@@ -1012,7 +1144,7 @@ def _construct_field(field, value):
             _c, _d = _construct_field(field.child_type, item)
             if _c is not None:
                 clean.append(_c)
-            if _d is not None:
+            if _d is not None and _d != "":
                 dropped.append(_d)
         return clean or None, dropped or None
 
@@ -1030,7 +1162,7 @@ def _construct_field(field, value):
             return None, value
 
 
-def construct_safe(mod, data) -> typing.Tuple[typing.Any, typing.Dict]:
+def construct_safe(mod, data) -> tuple[Any, dict]:
     if not isinstance(data, dict):
         return None, data
     fields = mod.fields()
